#include "WPILib.h"
#include "CANJaguar/CANJaguar.h"
#ifndef KICKER_H
#define KICKER_H
#define Power_Level_Tolerance 0.15 //Sets the tolerance (0-1) that the kicker "charging" accepts
#define Kicker_Frame_Perimeter 3.52 //This is the magnetic pot value that represents the 
								   //kicker being just inside the frame perimeter
#define Kicker_Fully_Charged 4.4 //This sets the fully reversed setting, used the magnetic pot value
// new fully charged is 0.005
#define Kicker_Offset 0 //This is the number between 0-1 that you need to subtract from the magpot
                        //to get it so that "0" is pointing straight out from the robot
#define Pulse_Rate 1 //Pulses the relays one in (Pulse_Rate) tics
#define Beater_Speed 0.3 //Speed the beater brush moves at
#define Fire_Delay .35 //Time it forces the EM to stay off
#define Prematch_Delay 3 //The number of seconds the kicker waits before it assumes it has unlatched if it hasn't registered it has



/*
 * 
 * VERY IMPORTANT:
 * WHENEVER MODIFYING THE MAGNETIC POTENTIOMETER:
 * MAKE SURE TO FIND THE FRAME PERIMETER VALUE
 * AND SET IT ABOVE THIS IN THE #define
 * IF YOU DO NOT DO THIS WE WILL GET PENTALTIES
 * 
 */



class Kicker {
public:
	Kicker();
	~Kicker();
	//Returns whether it has fired or not, and re-cocks it afterwards
	//Reasons why it may not immediately fire:
	//(If the sensor is not tripped, If it has not re-cocked yet)
	//(If 2 seconds have not passed since last kick)
	//(If the magic smoke has been released... D; )
	bool Fire();
	
	//Sets the power level to the float you pass it
	//Range is 0 to 1, 1 being the most powerful
	//IF POWER LEVEL SENSOR SYSTEM IS BROKEN:
	//No power level change takes place, all checks and
	//returns will assume that the power is what is desired
	void SetPowerDesired(float power);

	//Moves the kicker to the power level set in SetPowerDesired
	//Note: THIS MUST BE CALLED REPEATEDLY
	//You can NOT call this once and leave it. 
	void ProcessKicker();
	
	//Sets the beater brush's direction, attract, off, or repel
	//Pass a "1" for attract, a "0" for stop and a "-1" for repel
	//Any other values will cause it to stop (for safety reasons)
	void TurnBeaterBrush(int direction);
	
	//Checks to see if the power level is there or not
	//True = at the desired value (from SetPowerLevel)
	bool CheckPowerLevel();
	
	//Returns what the current power level is
	float GetPowerLevel();
	
	//Returns the raw pot value, THIS IS FOR DEBUGGING ONLY
	//USE GETPOWERLEVEL FOR ALL ROBOT PURPOSES
	float GetRawPot();
	
	//Returns the magpot value after calculations, THIS IS FOR DEBUGGING ONLY
	//USE GETPOWERLEVEL FOR ALL ROBOT PURPOSES
	float GetMagPot();
	
	//Returns what the desired power level is
	float GetPowerDesired();
	
	int GetForwardLimit();
	int GetBackwardLimit();
	int GetMagnetTouched();
	int GetRawBallSensor();
	
	//Checks to see if the shooter has waited the 2 seconds
	//DOES NOT CHECK IF THE POWER LEVEL IS WHERE IT SHOULD BE (See CheckPowerLevel)
	//DOES NOT CHECK IF A BALL IS AVAILABLE (See BallAvailable)
	bool ShooterReady();
	
	//Checks to see if the sensor detecting if a ball is in the shooter is tripped
	//Returns true if there is a ball ready to be kicked
	bool BallAvailable();
	
	//Returns if the sensor detecting balls in front of the robot is broken
	bool GetBallSensorBroken();
	
	//Returns if the sensor detecting the power level is broken
	bool GetPowerLevelSensorBroken();
	
	//Tells me if the ball sensor (banner sensor on the front) is broken or not
	//In case you couldn't guess: true = it's broken, false = it's not
	void SetBallSensorBroken(bool value);
	
	//Tells me if the upper mechanism sensors are broken
	//In case you couldn't guess: true = it's broken, false = it's not
	void SetPowerLevelSensorBroken(bool value);
	
	//Tells me that I should pull the kicker back to release the latch in processkicker
	//Lasts 2 seconds (More than enough time)
	void SetPreMatch();
	
	//Tells if the kicker has been releaed from its trigger lock
	bool GetPreMatchDone();
	
	//Resets if the Emagnet is broken (due to forward limit)
	void Reset();
	
private:
	bool FloatEquality(float test, float comparison, float tolerance);
	//Starts a timer to see when we can fire again (legaly). 
	//Use ShooterReady to see if the timer is up.
	void StartTimer();
	int relayPulse;
	double retractedTime;
	double preMatchTime;
	double fireTime;
	double timeoutTime;
	int LastSwitchState;
	float LastPotValue;
	float desiredPowerLevel;
	float kickerPosition;
	bool isTimeoutRunning;
	bool emagnetSwitchBroken;
	bool ballSensorBroken;
	bool powerLevelSensorBroken;
	bool kickerPreMatch;
	Relay * KickerMotor1;
	Relay * KickerMotor2;
	DigitalInput * BannerSensor;
	DigitalInput * MagnetTouched;
	DigitalInput * KickerFullForward;
	DigitalInput * KickerFullReverse;
	AnalogChannel * MagPot;
	Relay * EMagnet;
	Relay * EMagnet2;
	CANJaguar * BeaterBrush;
	int BeaterBrushDirection;
};
#endif

