/*----------------------------------------------------------------------------*/
/* Copyright (c) FIRST 2008. All Rights Reserved.							  */
/* Open Source Software - may be modified and shared by FRC teams. The code   */
/* must be accompanied by the FIRST BSD license file in $(WIND_BASE)/WPILib.  */
/*----------------------------------------------------------------------------*/

#include <vxWorks.h>

#include <errnoLib.h>
#include <hostLib.h>
#include <inetLib.h>
#include <sockLib.h>
#include <ioLib.h>
#include <ioctl.h>

#include "ThunderAxisCamera.h"
#include "Vision2009/BaeUtilities.h"
#include "Vision2009/FrcError.h"
#include "ChipObject/NiRioStatus.h"
#include "Task.h"
#include "Timer.h"
#include "Utility.h"

#include "ThunderVideoServer.h"

/**
 * @brief Implements an object that automatically does a close on a
 * camera socket on destruction.
 */
class ScopedSocket {
public:
	ScopedSocket(int camSock, const ErrorBase* errorBase)
		: m_camSock(camSock)
	{
		if (errorBase->StatusIsFatal())
		{
			return;
		}
		if (m_camSock == ERROR)
		{
			wpi_setError(*errorBase, ERR_CAMERA_SOCKET_CREATE_FAILED);
			return;
		}
	}

	~ScopedSocket() {
		if (m_camSock != ERROR) {
			close(m_camSock);
		}
	}
	// Cast to int allows you to pass this to any function that
	// takes the socket as an int.
	operator int() const {
		return m_camSock;
	}

private:
	int m_camSock;
};

//============================================================================
// ThunderVideoServer
//============================================================================

/**
 * @brief Constructor.
 */
ThunderVideoServer::ThunderVideoServer(void)
	: m_serverTask("ThunderVideoServer", (FUNCPTR)s_ServerTask)
	, m_newImageSem (NULL)
	, m_stopServer (false)
	, m_taskOver (false)
{
	m_newImageSem = ThunderAxisCamera::GetInstance()->GetNewImageSem();
	StartServerTask();
}

/**
 * @brief Destructor.
 * Stop serving images and destroy this class.
 */
ThunderVideoServer::~ThunderVideoServer()
{
	// Stop the images to PC server.
	Stop();
	// Clear the error so that you can use this object to make a connection to
	// the VIDEO_TO_PC_PORT to stop the ImageToPCServer if it is waiting to
	// accept connections from a PC.
	ClearError();
	// Wait until task is really finished
	while(!m_taskOver) {
		//Wait a bit so we don't take up all the processor time
		Wait(0.1);
	}
}

/**
 * Start the task that is responsible for sending images to the PC.
 */
int ThunderVideoServer::StartServerTask()
{
	if (StatusIsFatal())
	{
		return -1;
	}
	int id = 0;
	m_stopServer = false;
	// Check for prior copy of running task
	int oldId = taskNameToId((char*)m_serverTask.GetName());
	if(oldId != ERROR)
	{
		// TODO: Report error. You are in a bad state.
		taskDelete(oldId);
	}

	// spawn video server task
	// this is done to ensure that the task is spawned with the
	// floating point context save parameter.
	bool started = m_serverTask.Start((int)this);

	id = m_serverTask.GetID();

	if (!started)
	{
		wpi_setError(*this, ERR_CAMERA_TASK_SPAWN_FAILED);
		return id;
	}
	taskDelay(1);
	return id;
}

/**
 * @brief Start sending images to the PC.
 */
void ThunderVideoServer::Start()
{
	m_taskOver = false; // task has started over again
	StartServerTask();
}

/**
 * @brief Stop sending images to the PC.
 */
void ThunderVideoServer::Stop()
{
	m_stopServer = true;
}

/**
 * Static stub for kicking off the server task
 */
int ThunderVideoServer::s_ServerTask(ThunderVideoServer *thisPtr)
{
	return thisPtr->ServerTask();
}

/**
 * @brief Initialize the socket and serve images to the PC.
 * This is the task that serves images to the PC in a loop. This runs
 * as a separate task.
 */
int ThunderVideoServer::ServerTask()
{
	/* Setup to PC sockets */
	struct sockaddr_in serverAddr;
	int nonblock = TRUE;
	int sockAddrSize = sizeof(serverAddr);
	int pcSock = ERROR;
	bzero ((char *) &serverAddr, sockAddrSize);
	serverAddr.sin_len = (u_char) sockAddrSize;
	serverAddr.sin_family = AF_INET;
	serverAddr.sin_port = htons (VIDEO_TO_PC_PORT);
	serverAddr.sin_addr.s_addr = htonl (INADDR_ANY);

	int success;
	while (!m_stopServer)
	{
		taskSafe();
		// Create the socket.
		if ((pcSock = socket (AF_INET, SOCK_STREAM, 0)) == ERROR)
		{
			perror ("socket");
			taskUnsafe();
			continue;
		}
		// Set the TCP socket so that it can be reused if it is in the wait state.
		int reuseAddr = 1;
		setsockopt(pcSock, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&reuseAddr), sizeof(reuseAddr));
		ioctl(pcSock, FIONBIO, (int)&nonblock);
		// Bind socket to local address.
		if (bind (pcSock, (struct sockaddr *) &serverAddr, sockAddrSize) == ERROR)
		{
			perror ("bind");
			close (pcSock);
			taskUnsafe();
			continue;
		}
		// Create queue for client connection requests.
		if (listen (pcSock, 1) == ERROR)
		{
			perror ("listen");
			close (pcSock);
			taskUnsafe();
			continue;
		}

		struct sockaddr_in clientAddr;
		int clientAddrSize;
		int newPCSock;
		do {
			taskUnsafe();
			Wait(0.1);
			taskSafe();
			newPCSock = accept (pcSock, reinterpret_cast<sockaddr*>(&clientAddr), &clientAddrSize);
		} while (newPCSock == ERROR && errno == EWOULDBLOCK && !m_stopServer);
		if (newPCSock == ERROR || m_stopServer)
		{
			close(pcSock);
			taskUnsafe();
			continue;
		}
		//TODO: check camera error

		int numBytes = 0;
		int imageDataSize = 0;
		char* imageData = NULL;

		while(!m_stopServer)
		{
			success = semTake(m_newImageSem, 1000);
			if (success == ERROR)
			{
				// If the semTake timed out, there are no new images from the camera.
				continue;
			}
			success = ThunderAxisCamera::GetInstance()->CopyJPEG(&imageData, numBytes, imageDataSize);
			if (!success)
			{
				// No point in running too fast -
				Wait(1.0);
				// If camera is not initialzed you will get failure and
				// the timestamp is invalid. Reset this value and try again.
				continue;
			}

			// Write header to PC
			static const char header[4]={1,0,0,0};
			int headerSend = write(newPCSock, const_cast<char*>(header), 4);

			// Write image length to PC
			int lengthSend = write(newPCSock, reinterpret_cast<char*>(&numBytes), 4);

			// Write image to PC
			int sent = write (newPCSock, imageData, numBytes);

			// The PC probably closed connection. Get out of here
			// and try listening again.
			if (headerSend == ERROR || lengthSend == ERROR || sent == ERROR)
			{
				break;
			}
		}
		// Clean up
		delete [] imageData;
		close (newPCSock);
		newPCSock = ERROR;
		close (pcSock);
		pcSock = ERROR;
		taskUnsafe();
		Wait(0.1);
	}
	m_taskOver = true;
	return (OK);
}


